/* Trivial program to run conformance tests for cmd.exe
 * on either Windows or Wine
 * LGPL, for inclusion in Wine source tree
 * Copyright 2009, Dan Kegel
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>

/* generated by cmdtest-pack.sh */
#include "cmdtest-data.h"

/*
 * Exit with status zero if on Win9X where cmd.exe is not supported.
 * TODO: support win9x and batch files someday.
 */
void detect_os(void)
{
    const char *os;
    HANDLE ntdll;

    ntdll = GetModuleHandleA("ntdll.dll");
    if (ntdll) {
        if (GetProcAddress(ntdll, "wine_get_build_id")) {
            /* Supported OS */
            return;
        }
    }

    os = getenv("OS");
    if (os && !strcmp(os, "Windows_NT")) {
        /* Supported OS */
        return;
    }

    fprintf(stderr, "Unsupported operating system, skipping cmd tests\n");
    exit(0);
}

static void cmdtest_unpack()
{
   /* Unpack the data files.  FIXME: replace embedded arrays with resources. */
   int i;
   for (i=0; cmdtest_files[i].name; i++) {
      FILE *fp = fopen(cmdtest_files[i].name, "r");
      if (fp) {
         fprintf(stderr, "%s already exists!\n", cmdtest_files[i].name);
         exit(1);
      }

      fp = fopen(cmdtest_files[i].name, "wb");
      int len;
      int nwritten;
      if (!fp) {
         perror(cmdtest_files[i].name);
         exit(1);
      }
      len = strlen(cmdtest_files[i].data);
      nwritten = fwrite(cmdtest_files[i].data, 1, len, fp);
      if (len != nwritten) {
         perror(cmdtest_files[i].name);
         exit(1);
      }
      fclose(fp);
   }
}

/* Compare given output to reference, return 0 on identical, nonzero on different */
/* FIXME: don't quit after first error; skip to next test in file */
static int check(const char *name, const char *suffix)
{
    char fname_out[2048];
    char fname_exp[2048];
    FILE *fp_out = NULL, *fp_exp = NULL;
    char buf_out[2048];
    char buf_exp[2048];
    int line;
    int ret = 1;

    sprintf(fname_out, "%s.%s", name, suffix);
    fp_out = fopen(fname_out, "rb");

    sprintf(fname_exp, "%s.%s.exp", name, suffix);
    fp_exp = fopen(fname_exp, "rb");

    if (fp_exp && ! fp_out) {
        fprintf(stderr, "%s exists, but %s doesn't\n", fname_exp, fname_out);
        goto out;
    }
    if (!fp_exp && fp_out) {
        fprintf(stderr, "%s exists, but %s doesn't\n", fname_out, fname_exp);
        goto out;
    }

    line = 0;
    while (1) {
        const char *pout, *pexp;
        line++;
        pout = fgets(buf_out, sizeof(buf_out), fp_out);
        pexp = fgets(buf_exp, sizeof(buf_exp), fp_exp);
        if (!pout && !pexp)
            break;

        if (pexp && ! pout) {
            fprintf(stderr, "%s EOF before %s\n", fname_out, fname_exp);
            goto out;
        }
        if (!pexp && pout) {
            fprintf(stderr, "%s EOF before %s\n", fname_exp, fname_out);
            goto out;
        }
        if (strcmp(pout, pexp)) {
            fprintf(stderr, "%s and %s differ on line %d: '%s' vs '%s'\n", fname_out, fname_exp, line, pout, pexp);
            goto out;
        }
    }
    ret= 0;
out:
    if (fp_out)
        fclose(fp_out);
    if (fp_exp)
        fclose(fp_exp);
    printf("check %s %s returning %d\n", name, suffix, ret);
    return ret;
}

/* Return 0 on success, nonzero on failure */
static int run_test(const char *name)
{
    DWORD err;
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    DWORD wait, status;
    HANDLE hOut;
    char outname[2048];
    HANDLE hErr;
    char errname[2048];
    char tempdir[2048];
    char cmd[2048];

    GetTempPath(sizeof(tempdir), tempdir);

    sprintf(outname, "%s.out", name);
    hOut = CreateFileA(outname,
				FILE_WRITE_DATA,
				FILE_SHARE_READ,
				NULL,          
				CREATE_NEW,  
				FILE_ATTRIBUTE_NORMAL,
				NULL);

    //WriteFile(hOut, "howdy\n", 6, &err, 0);

    sprintf(errname, "%s.err", name);
    hErr = CreateFileA(errname,
				FILE_WRITE_DATA,
				FILE_SHARE_READ,
				NULL,          
				CREATE_NEW,  
				FILE_ATTRIBUTE_NORMAL,
				NULL);

    GetStartupInfo (&si);
    si.dwFlags    = STARTF_USESTDHANDLES;
    si.hStdInput  = GetStdHandle( STD_INPUT_HANDLE );
    si.hStdOutput = hOut;
    si.hStdError  = hErr;
 
    sprintf(cmd, "cmd /c echo foo %s.cmd", name);
    if (!CreateProcessA (NULL, cmd, NULL, NULL, TRUE, CREATE_DEFAULT_ERROR_MODE,
                         NULL, NULL /*tempdir*/, &si, &pi))
        return -2;

    CloseHandle (pi.hThread);
    if (WaitForSingleObject(pi.hProcess, INFINITE))
        printf("wait failed?\n");

    err = 1234;
    if (!GetExitCodeProcess(pi.hProcess, &err))
        printf("GetExitCodeProcess failed?\n");

    if (err != 0)
        printf("running %s.cmd returns exit status %d\n", name, err);

    CloseHandle(hOut);
    CloseHandle(hErr);

    if (err == 0 && !check(name, "out") && !check(name, "err"))
      return 0;
    return 1;
}

int main(int argc, char **argv)
{
   int failures;
   int i;

   detect_os();
   cmdtest_unpack();

   /* Suppress warns and fixmes */
   putenv("WINEDEBUG=-all");

   failures=0;
   for (i=0; cmdtest_files[i].name; i++) {
      const char *fname = cmdtest_files[i].name;
      const char *dot = fname + strlen(fname) - 4;
      char name[1024];

      /* Skip everything but .cmd files */
      if (strcmp(dot, ".cmd"))
          continue;

      /* make base name without suffix */
      memcpy(name, fname, dot - fname);
      name[dot - fname] = 0;

      if (run_test(name)) {
         printf("test %s failed\n", name);
         failures++;
      }
   }
   if (failures == 0) {
       printf("All tests passed.\n");
       return 0;
   }
   printf("%d tests failed.\n", failures);
   return 1;
}
